
// Jau.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, March 2014

/* Jau is a set of static methods to make it easy for a
   Java program to invoke a separate MS Windows application, and communicate
   with it via its GUI interface.

   This is a rather 'fragile' way to interact with other applications, since the 
   behaviour of a GUI interface can be hard to reliably control. For example, if the user
   clicks on another window, the 'focus' of the OS will change, and Java will lose the
   link to the original application. Also, the behaviour of a GUI, especially when an application first
   starts, can be hard to predict. For example, a "Do you Want to Update this Software" window
   might appear, or pop-up message windows could appear at different intervals. The appearance of
   these windows will change the active window, and again Java will lose its link to the 
   application.

   The main reason for using this approach is that the application in question does not offer
   a Java-based API.

   Jau contains several groups of methods:

     * application location methods: these are methods for listing out the applications
       known to Windows, which mean they can be called without supplying a long path name
       to the executable. See TestApps.java for an example of their use.

     * run methods: these are methods for starting another application, and receiving back a 
       windows handle that (of type HWND) which can be used for subsequent communication.
       See TestRun.java for examples.

     * windows methods: these are used to control the state, size, position of the window. 
       Common states include making the window active, or minimizing it. Another important 
       feature is being able to read and write text to the window.
       See TestWin.java and TestDesktop.java.

     * controls methods: most application are not just a text window, but utilize GUI controls
       such as buttons, menus, lists and tree views, the status bar, and clipboard. 
       These methods allow these controls to be
       examined and changed. The main way of doing this is by using the enclosing window's handle
       and the control's ID. This ID can be found using a tool such as WinSpy++ (http://www.catch22.net/software/winspy-17) for examining the properties of windows.
       See the various "Use" examples.

     * process methods: these methods allow things like the process ID (pid) of a window to
       be examined. They will probably not be needed by a user of Jau who can manage a window
       using it's handle alone.

     * mouse methods: these methods are also unlikely to be needed, since it's more reliable to 
       access the controls in a window via their control IDs.


   Jau is built on top of a JNA wrapper around the AutoIt DLL for automating access to Windows,
   which is defined in AutoItX3.java. Please see the comments in that file for details on how
   the wrapper is implemented.

   The main contribution of Jau is to simplify the access to Windows offered by AutoIt. This
   includes the simplification of the signatures of many methods, removing (what I believe)
   to be unnecessary arguments, and replacing a C-like parameter style by something more 
   Java-like. 

   Bearing in mind that I might be wrong in my simplification choices, the original
   AutoIt functions are still available, but the user will have to call them via AutoItX3
   (see UseToolTip.java for an example).
*/


import java.io.*;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Insets;

import java.util.*;
import javax.swing.JOptionPane;
import java.awt.datatransfer.*;

import com.sun.jna.*;
import com.sun.jna.win32.*;


import com.sun.jna.platform.win32.*;
import com.sun.jna.platform.win32.WinDef.*;
import com.sun.jna.platform.win32.WinUser.*;



public class Jau
{
  public static final int MAX_RUN_WAIT = 5000;   // ms, used by run()

  public static final int BUF_SIZE = 255;


  // used when setting window display functions, as in winSetState()
  // from http://msdn.microsoft.com/en-us/library/windows/desktop/ms633548(v=vs.85).aspx
  public static final int SW_HIDE = 0;   
  public static final int SW_SHOW = 5;
  public static final int SW_MAXIMIZE = 3; 
  public static final int SW_MINIMIZE = 6; 
  public static final int SW_RESTORE = 9;
  public static final int SW_ENABLE = 64;
  public static final int SW_DISABLE = 65;

  public static final int SW_SHOWNORMAL = 1;     
  public static final int SW_SHOWMINIMIZED = 2;     
  public static final int SW_SHOWMAXIMIZED = 3;  
  public static final int SW_SHOWNOACTIVATE = 4;
  public static final int SW_SHOWMINNOACTIVE = 7;
  public static final int SW_SHOWNA = 8;
  public static final int SW_SHOWDEFAULT = 10;
  public static final int SW_FORCEMINIMIZE = 11;


  // used when testing window state, as in winGetState()
  public static final int WIN_EXISTS = 1;
  public static final int WIN_VISIBLE = 2;
  public static final int WIN_ENABLED = 4;
  public static final int WIN_ACTIVE = 8;
  public static final int WIN_MINIMIZED = 16;
  public static final int WIN_MAXIMIZED = 32;



  // mouse button labels, used in controlClick()
  public static final String BTN_LEFT = "left";
  public static final String BTN_RIGHT = "right";
  public static final String BTN_MIDDLE = "middle";
  public static final String BTN_MAIN = "main";
  public static final String BTN_MENU = "menu";
  public static final String BTN_PRIMARY = "primary";
  public static final String BTN_SECONDARY = "secondary";

  public static final String[] BTNS = { "left", "right", "middle", "main",
                                        "menu", "primary", "secondary" };


  // mouse wheel labels
  public static final String WHEEL_UP = "up";
  public static final String WHEEL_DOWN = "down";

  public static final String[] DIRS = { "up", "down" };



  // window title match modes; see winTitleMatchMode()
  // see http://www.autoitscript.com/autoit3/docs/functions/AutoItSetOption.htm
  public static final int MATCH_FROM_START = 1;   // default
  public static final int MATCH_ANY = 2;   
  public static final int MATCH_EXACT = 3;   
  public static final int MATCH_ADV = 4;

  public static final int MATCH_FROM_START_NO_CASE = -1;
  public static final int MATCH_ANY_NO_CASE = -2;   
  public static final int MATCH_EXACT_NO_CASE = -3;   
  public static final int MATCH_ADV_NO_CASE = -4;


  // window caret coord modes; see winCaretCoordMode()
  // see http://www.autoitscript.com/autoit3/docs/functions/AutoItSetOption.htm
  public static final int CC_WINDOW = 0;   // relative coords to the active window
  public static final int CC_SCREEN = 1;   // (default) absolute screen coordinates
  public static final int CC_CLIENT = 2;   
               // relative coords to the client area of the active window


  // mouse coord modes; see mouseCoordMode()
  // see http://www.autoitscript.com/autoit3/docs/functions/AutoItSetOption.htm
  public static final int M_WINDOW = 0;   // relative coords to the active window
  public static final int M_SCREEN = 1;   // (default) absolute screen coordinates
  public static final int M_CLIENT = 2;   
               // relative coords to the client area of the active window




  public interface  User32Ext extends User32
  {
    // Create an extended JNA proxy for user32.dll ...
    User32Ext INSTANCE = (User32Ext) Native.loadLibrary("user32", User32Ext.class,
                                        W32APIOptions.DEFAULT_OPTIONS);

   // docs at http://twall.github.io/jna/3.4.0/javadoc/com/sun/jna/platform/win32/User32.html


    // ---------------- additional methods --------------

    // for SetWindowPosition()
    int SWP_NOSIZE = 0x0001;
    int SWP_NOMOVE = 0x0002;

    WinDef.HWND HWND_TOP = new WinDef.HWND(Pointer.createConstant(0));
    WinDef.HWND HWND_BOTTOM = new WinDef.HWND(Pointer.createConstant(1));


    int CascadeWindows(HWND hwnd, int how, RECT rect, int kids, HWND KidHandles);

    boolean EnableWindow(HWND handle, boolean isEnabled);

    // boolean GetCaretPos(WinDef.POINT point);

    int GetDlgCtrlID(HWND hWnd);

    HWND GetDesktopWindow();

    HWND GetTopWindow(HWND hWnd);

    boolean IsWindow(HWND handle);

    boolean IsWindowEnabled(HWND handle);

    int TileWindows(HWND hwnd, int how, RECT rect, int kids, HWND kidsHandles);

    // HWND WindowFromPoint(WinDef.POINT Point);
    HWND WindowFromPoint(int xPoint, int yPoint);
          // https://groups.google.com/forum/#!topic/jna-users/jVSfNCHMQ-M

  }  // end of User32Ext interface

  public static User32Ext user32ext = User32Ext.INSTANCE;




  // Create a JNA proxy for kernel32.dll ...
  public static Kernel32 kernel32 = (Kernel32) Native.loadLibrary("kernel32", Kernel32.class,
                                        W32APIOptions.DEFAULT_OPTIONS);



  public interface PsAPI extends StdCallLibrary
  // Use one method from psapi.dll
  {
    PsAPI INSTANCE = (PsAPI) Native.loadLibrary("Psapi", PsAPI.class,
                                        W32APIOptions.DEFAULT_OPTIONS);

    // http://msdn.microsoft.com/en-us/library/ms683198(VS.85).aspx
    int GetModuleFileNameExA(WinNT.HANDLE process, WinNT.HANDLE module,
                                                        byte[] fileName, int size);
  } // end of PsAPI interface

  public static PsAPI psapi = PsAPI.INSTANCE;




  // ======================= start of Jau methods ============================



  public static int error()
  // get the error result of the last function call
  {  return AutoItX3.INSTANCE.AU3_error(); }


  public static int autoItSetOption(String option, int param)
  // http://www.autoitscript.com/autoit3/docs/functions/AutoItSetOption.htm
  {   return AutoItX3.INSTANCE.AU3_AutoItSetOption(option, param);  }




  // ------------------------- application location methods ---------------------------------
  //  (not in AutoItX)




  public static String appFindPath(String appName)
  // return a single app path by searching the OSes known applications
  {
    List<String> appPaths = appListPaths(appName);
    if (appPaths.size() == 0) {
      // System.out.println("No applications match \"" + appName + "\"");
      return null;
    }

    if (appPaths.size() > 1) {
      System.out.println(appPaths.size() + " applications match \"" + appName + "\"");
      System.out.println("Returning the first: \"" + appPaths.get(0) + "\"");
    }

    return appPaths.get(0);
  }  // end of appFindPath()



  public static List<String> appListPaths(String appName)
  // return all matching app paths by searching the OSes known applications
  {
    List<String> allAppPaths = appListPaths();
    List<String> matchPaths = new ArrayList<String>();
    String nm = appName.toLowerCase();

    for(String path : allAppPaths) {
      if (path.toLowerCase().contains(nm)) {
        // System.out.println("appListPaths: " + path);
        matchPaths.add(path);
      }
    }
    return matchPaths;
  } // end of appListPaths()




  public static List<String> appListPaths()
  /* return all the application paths recorded in the appPath 
     registry key 
  */
  {
    List<String> exePaths = new ArrayList<String>();
    String appPathsKey = "Software\\Microsoft\\Windows\\CurrentVersion\\App Paths";
    String[] apkeys = Advapi32Util.registryGetKeys(
                            WinReg.HKEY_LOCAL_MACHINE, appPathsKey);
    for (String key : apkeys) {
      // System.out.println(key);
      String exePath = registryGet(
                                  WinReg.HKEY_LOCAL_MACHINE,
                                  appPathsKey + "\\" + key, "");
      if (exePath != null) {
        // System.out.println("appListPaths: " + exePath);
        exePaths.add(exePath);
      }
    }
    return exePaths;
  }  // end of appListPaths()



  public static String registryGet(WinReg.HKEY root, String key, String value)
  { try {
      return Advapi32Util.registryGetStringValue(root, key, value);
    }
    catch(Win32Exception e)
    {  // System.out.println("registry get failed");
       return null;
    }
  }  // end of registryGet()




  public static String appsDir()
  {  return registryGet(WinReg.HKEY_LOCAL_MACHINE, 
             "SOFTWARE\\Microsoft\\Windows\\CurrentVersion", "ProgramFilesDir");  }


  public static String userDir()
  {  return System.getProperty("user.dir");  }


  public static String desktopPath()
  {  return (System.getProperty("user.home") + "\\Desktop");  }


  // -------------------------  run methods ----------------------------------



  public static HWND run(String appName)
  {  return run(appName, MAX_RUN_WAIT);  }


  public static HWND run(String appName, int timeOut)
  /* Runs an application.
     Returns the window handle after waiting for the application to start executing.
  */
  {
    // invoke the appName application, getting its process ID
    int pid = AutoItX3.INSTANCE.AU3_Run(appName, ".", SW_SHOWNORMAL);
    if (pid == 0) {   // the OS run() does not know appName
      String appPath = appFindPath(appName);   // search through OS app Name list, try appPath instead
      if (appPath != null) {
        System.out.println("Calling \"" + appPath + "\"");
        appName = appPath;
        pid = AutoItX3.INSTANCE.AU3_Run(appName, ".", SW_SHOWNORMAL);
      }
      else {
        System.out.println("Sorry, could not find a matching application to run for \"" + appName + "\"");
        return null;
      }
    }

    // wait for an executing window to appear with the pid value
    HWND hWnd = processGetHandle(pid);
    if (hWnd == null)
      System.out.print("Waiting for \"" + appName + "\" to be detected...");

    int counter = 0;
    while ((hWnd == null) && (counter < timeOut)) {
      sleep(1000);
      System.out.print(".");
      hWnd = processGetHandle(pid);
      counter += 1000;
    }
    System.out.println();

    // return the window's handle
    if (hWnd == null)
      System.out.println("\"" + appName + "\" not detected; try adding a title argument to run()");
    else
      System.out.println("Handle: " + Jau.handleString(hWnd));

    return hWnd;
  }  // end of run()



  public static HWND run(String appName, String title)
  {  return run(appName, title, MAX_RUN_WAIT);  }



  public static HWND run(String appName, String title, int timeOut)
  /* Runs an application.
     Returns the window handle of the titled window after the application as
     started. 
  */
  {
    // invoke the appName application, getting its process ID
    int pid = AutoItX3.INSTANCE.AU3_Run(appName, ".", SW_SHOWNORMAL);
    if (pid == 0) {    // the OS run() does not know appName
      String appPath = appFindPath(appName);   // search through OS app Name list, try appPath instead
      if (appPath != null) {
        System.out.println("Calling \"" + appPath + "\"");
        appName = appPath;
        pid = AutoItX3.INSTANCE.AU3_Run(appName, ".", SW_SHOWNORMAL);
      }
      else {
        System.out.println("Sorry, could not find a matching application to run for \"" + appName + "\"");
        return null;
      }
    }

    // wait for an executing window to appear with the specified title
    System.out.println("Waiting for \"" + title + "\" to be detected...");
    boolean isActive = winWaitActive(title, timeOut);
    if (!isActive) {
      System.out.println("\"" + title + "\" not detected");
      return null;
    }

    HWND hWnd = winGetHandle(title);
    System.out.println("Handle: " + Jau.handleString(hWnd));
    return hWnd;
  }  // end of run()




  // --------------------------process management ----------------------------
  // the methods use a PID integer rather than a process name in AutoIt


  public static HWND processGetHandle(int pid)
  /* check pid against all the pid of executing windows to find the
     corresponding window handle  (not in AutoItX) */
  {
    List<HWND> hWindows = winListHandles();
    for(HWND hWnd : hWindows) {
      int winPID = winGetProcess(hWnd);
      // System.out.println("Window \"" + winGetTitle(hWnd) + "\" pid: " + winPID);
      if (pid == winPID) {
        // System.out.println("PID " + pid + " window handle is " + handleString(hWnd));
        return hWnd;
      }
    }
    return null;
  }  // end of processGetHandle()




  public static List<Integer> processListPIDs()
  /* return a list of system-wide process IDs;
     not in AutoItX, although something similar in function reference
  */
  { Map<Integer,String> map = processMapApps();
    return new ArrayList<Integer>(map.keySet());
  }


  public static List<Integer> processListPIDs(String appName)
  // return a list of PIDs associated with this appName (not in AutoItX)
  {  
    Map<Integer,String> map = processMapApps();
    List<Integer> pids = new ArrayList<Integer>();
    String nm = appName.toLowerCase();

    for (Map.Entry<Integer,String> entry : map.entrySet()) {
      if ((entry.getValue()).toLowerCase().contains(nm))
        pids.add( entry.getKey() );
    }
    return pids;
  }  // end of processListPIDs()


  public static Map<Integer,String> processMapApps()
  // generate a system-wide map of process IDs to app names (not in AutoItX)
  // http://stackoverflow.com/questions/9877993/jna-query-windows-processes
  {
    Map<Integer,String> pids = new HashMap<Integer,String>();
    //Kernel32 kernel32 = (Kernel32) Native.loadLibrary(Kernel32.class,
    //                                                W32APIOptions.UNICODE_OPTIONS);

    Tlhelp32.PROCESSENTRY32.ByReference processEntry = new Tlhelp32.PROCESSENTRY32.ByReference();
         //  a pointer to a PROCESSENTRY32 structure

    WinNT.HANDLE snapshot = kernel32.CreateToolhelp32Snapshot(
                                  Tlhelp32.TH32CS_SNAPPROCESS, new WinDef.DWORD(0));

    try {
      while (kernel32.Process32Next(snapshot, processEntry)) {
        /* get next pointer to entry from a list of the processes residing in the
           system address space when a snapshot was taken */
        int pid = (processEntry.th32ProcessID).intValue();
        String appName = Native.toString(processEntry.szExeFile);
        // System.out.println(pid + "\t" + appName );
        pids.put(pid, appName);
      }
    }
    finally {
      kernel32.CloseHandle(snapshot);
    }

    return pids;
  } // end of processMapApps()



  public static String processGetApp(int pid)
  // map a process ID to an app name (not in AutoItX)
  { Map<Integer,String> map = processMapApps();
    return map.get(pid);
  }



  // ---------- methods that apply to multiple running windows ----------------



  public static HWND appGetHandle(String appName)
  // get window handle for running app (if there is one)
  {
    String appPath = appFindPath(appName);
    if (appPath == null)
      return null;

    List<HWND> handles = new ArrayList<HWND>();
    Map<HWND,String> map = winMapApps();
    for (Map.Entry<HWND,String> entry : map.entrySet()) {
      if (appPath.equals(entry.getValue()))
        handles.add(entry.getKey());
    }

    if (handles.size() == 0) {
      // System.out.println("No applications match \"" + appName + "\"");
      return null;
    }

    if (handles.size() > 1) {
      System.out.println(handles.size() + " applications match \"" + appName + "\"");
      System.out.println("Returning handle of the first");
    }

    return handles.get(0);
  }  // end of appGetHandle()



  public static List<String> winListApps(String appName)
  // return all matching app paths
  {
    List<String> winAppPaths = winListApps();
    List<String> matchPaths = new ArrayList<String>();
    String nm = appName.toLowerCase();

    for(String path : winAppPaths) {
      if (path.toLowerCase().contains(nm)) {
        // System.out.println("appListPaths: " + path);
        matchPaths.add(path);
      }
    }
    return matchPaths;
  } // end of winListApps()



  public static List<String> winListApps()
  // list of application paths for all the executing windows
  {
    Map<HWND,String> map = winMapApps();
    return new ArrayList<String>(map.values());
  }



  public static Map<HWND,String> winMapApps()
  /* a map of window handles to application path names for
     executing windows (not in AutoIt)  */
  {
    List<HWND> hWindows = winListHandles();
    Map<HWND,String> winMapApps = new HashMap<HWND, String>();
    String appName;
    for(HWND hWnd : hWindows) {
      appName = winGetApp(hWnd);
      if ((appName != null) && (appName.length() > 0))
        winMapApps.put(hWnd, appName);
    }
    return winMapApps;
  }  // end of winMapApps()





  public static List<HWND> winListHandles()
  // return a list of handles for visible windows with titles (not in AutoIt)
  {
    final List<HWND> hWindows = new ArrayList<HWND>();

    user32ext.EnumWindows( new WNDENUMPROC()
    {
      public boolean callback(HWND hWnd, Pointer userData)
      /* called for each window found by EnumWindows().
         hWnd parameter is the window's handle */
      {
        if (!winVisible(hWnd))
          return true;

        String wTitle = winGetTitle(hWnd);
        // System.out.println("winListHandles title: " + wTitle);
        if ((wTitle != null) && (wTitle.length() != 0))
          hWindows.add(hWnd);

        return true;
      }
    }, null);

    // System.out.println("winListHandles length: " + hWindows.size());
    return hWindows;
  }  // end of winListHandles()



  public static List<HWND> winListHandles(String titlePart)
  // return a list of titles that contain titlePart
  {
    List<String> winListTitles = winListTitles();
    List<HWND> matchHandles = new ArrayList<HWND>();
    String tp = titlePart.toLowerCase();
    for(String title : winListTitles) {
      if (title.toLowerCase().contains(tp))    // more general matching than AutoIt
        matchHandles.add( winGetHandle(title));
    }
    return matchHandles;
  }  // end of winListHandles()




  public static List<String> winListTitles()
  // return a list of titles for visible windows
  {
    List<HWND> hWindows = winListHandles();
    List<String> winListTitles = new ArrayList<String>();
    for(HWND hWnd : hWindows)
      winListTitles.add( winGetTitle(hWnd));
    return winListTitles;
  }  // end of winListTitles()



  public static void winMinimizeAll()
  // http://www.autoitscript.com/autoit3/docs/functions/WinMinimizeAll.htm
  {  AutoItX3.INSTANCE.AU3_WinMinimizeAll();  }


  public static void winMinimizeAllUndo()
  // http://www.autoitscript.com/autoit3/docs/functions/WinMinimizeAllUndo.htm
  {  AutoItX3.INSTANCE.AU3_WinMinimizeAllUndo();  }



  public static int winCascadeAll()
  // cascade all the windows on the desktop
  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms632674%28v=vs.85%29.aspx
  // not in AutoIt
  {  return user32ext.CascadeWindows(null, 1, null, 0, null);  }



  public static int winTileAll()
  // horizontally tile all the windows on the desktop
  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms633554%28v=vs.85%29.aspx
  // not in AutoIt
  {  return user32ext.TileWindows(null, 1, null, 0, null);  }



  // ---------------------- methods used on a single window ----------------------------


  public static String handleString(HWND hWnd)
  // returns the handle as a hexadecimal string
  {  
    if (hWnd == null)
      return null;
    else
      return Long.toHexString(Pointer.nativeValue(hWnd.getPointer()));
  }


  public static boolean isWindow(String title)
  // not in AutoIt
  {  return user32ext.IsWindow( winGetHandle(title));  }

  public static boolean isWindow(HWND hWnd)
  {  return user32ext.IsWindow(hWnd);  }


  public static HWND winActivateBottom()
  // not in AutoIt
  {
    List<HWND> hWindows = winListHandles();
    if (hWindows == null)
      return null;
    else {
      HWND hWnd = hWindows.get( hWindows.size()-1);
      winActivate(hWnd);
      return hWnd;
    }
  }  // end of winActivateBottom()




  public static int winActivate(String title)
  // Activates (gives focus to) a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinActivate.htm
  {  return AutoItX3.INSTANCE.AU3_WinActivate(title, "");  }
     
  public static int winActivate(HWND hWnd)
  {  return AutoItX3.INSTANCE.AU3_WinActivateByHandle(hWnd);  }



  public static boolean winActive(String title)
  // Checks to see if a specified window exists and is currently active
  // http://www.autoitscript.com/autoit3/docs/functions/WinActive.htm
  {  return ((winGetState(title) & WIN_ACTIVE) == WIN_ACTIVE);  }
//   {  return (AutoItX3.INSTANCE.AU3_WinActive(title) == 1);  }

  public static boolean winActive(HWND hWnd)
  {  return ((winGetState(hWnd) & WIN_ACTIVE) == WIN_ACTIVE);  }
//   {  return (AutoItX3.INSTANCE.AU3_WinActiveByHandle(hWnd) == 1);  }



  public static int winCaretCoordMode(int flag)
  /* Sets the way coords are used in the caret functions, either absolute coords or 
     coords relative to the current active window:
         0 (or CC_WINDOW) = relative coords to the active window
         1 (or CC_SCREEN) = (default) absolute screen coordinates
         2 (or CC_CLIENT)= relative coords to the client area of the active window
    see http://www.autoitscript.com/autoit3/docs/functions/AutoItSetOption.htm
  */
  {  
    if ((flag >= 0) && (flag <= 3)) {
      System.out.println("Changing caret coord mode to " + flag);
      return autoItSetOption("CaretCoordMode", flag);
    }
    else {
      System.out.println("Illegal caret coord mode value: " + flag);
      return 0;
    }
  }  // end of winCaretCoordMode()




  public static boolean winClose(String title)
  // http://www.autoitscript.com/autoit3/docs/functions/WinClose.htm
  { System.out.println("Closing \"" + title + "\" (handle: " + 
              handleString(winGetHandle(title)) + ")");
    return (AutoItX3.INSTANCE.AU3_WinClose(title, "") == 1); 
  }

  public static boolean winClose(HWND hWnd)
  {
    System.out.println("Closing \"" + winGetTitle(hWnd) + "\" (handle: " +
                           handleString(hWnd) + ")"  );
    return (AutoItX3.INSTANCE.AU3_WinCloseByHandle(hWnd) == 1);
  }



  public static boolean winEnable(String title, boolean isEnabled)
  // 'enabled' means that the window can be interact with (not in AutoIt)
  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms646291(v=vs.85).aspx
  {  return user32ext.EnableWindow( winGetHandle(title), isEnabled);  }

  public static boolean winEnable(HWND hWnd, boolean isEnabled)
  {  return user32ext.EnableWindow(hWnd, isEnabled);  }


  public static boolean winEnabled(String title)
  {  return user32ext.IsWindowEnabled( winGetHandle(title));  }

  public static boolean winEnabled(HWND hWnd)
  {  return user32ext.IsWindowEnabled(hWnd);  }



  public static boolean winExists(String title)
  // http://www.autoitscript.com/autoit3/docs/functions/WinExists.htm
  {  return (AutoItX3.INSTANCE.AU3_WinExists(title, "") == 1);  }

  public static boolean winExists(HWND hWnd)
  {  return (AutoItX3.INSTANCE.AU3_WinExistsByHandle(hWnd) == 1);  }





  public static boolean winFlash(String title)
  //  flashes the specified window (not in AutoIt)
  // http://www.autoitscript.com/autoit3/docs/functions/WinFlash.htm
  // http://stackoverflow.com/questions/2773364/make-jface-window-blink-in-taskbar-or-get-users-attention
  {
    HWND hWnd = winGetHandle(title);
    if (hWnd == null)
      return false;
    else
      return winFlash(hWnd);
  }


  public static boolean winFlash(HWND hWnd)
  {
    User32.FLASHWINFO info = new User32.FLASHWINFO();
    info.cbSize = info.size();
    info.dwFlags = User32.FLASHW_ALL | User32.FLASHW_TIMERNOFG;
       // flash until the window comes to the foreground
    info.hWnd = hWnd;
    return user32ext.FlashWindowEx(info);
  }  // end of winFlash()



  public static String winGetApp(String title)
  { HWND hWnd = winGetHandle(title);
    return winGetApp(hWnd);
  }


  public static String winGetApp(HWND hWnd)
  // get the name of the executable that created this window (not in AutoIt)
  {
    int pid = winGetProcess(hWnd);
    // System.out.println("PID: " + pid);

    WinNT.HANDLE process = kernel32.OpenProcess(0x0400 | 0x0010, false, pid);
      // http://msdn.microsoft.com/en-us/library/ms684880(v=VS.85).aspx
        // kernel32.OpenProcess(0x1040, false, pid);

    if (process == null) {
      System.out.println("Could not access process info");
      return null;
    }

    byte[] name = new byte[1024];
    int result = psapi.GetModuleFileNameExA(process, null, name, 1024);

    String appName = Native.toString(name).substring(0, result).trim();
    // System.out.println("App Name: \"" + appName + "\"");

    return appName;
  }  // end of winGetApp()



  public static boolean winIsApp(String title, String name)
  { HWND hWnd = winGetHandle(title);
    return winIsApp(hWnd, name);
  }

  public static boolean winIsApp(HWND hWnd, String name)
  // does the window's application name contain the name string?
  {
    String appName = winGetApp(hWnd);
    if (appName == null)
      return false;

    return appName.contains(name);
  }  // end of winIsApp() 




  public static Point winGetCaretPos()
  // Returns the coordinates of the caret in the foreground window
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetCaretPos.htm
  { WinDef.POINT point = new WinDef.POINT();
    AutoItX3.INSTANCE.AU3_WinGetCaretPos(point);
    return new Point(point.x, point.y);
  }




  public static int winGetClientSizeWidth(HWND hWnd)
  {  return (winGetClientSize(hWnd)).width;  }

  public static int winGetClientSizeHeight(HWND hWnd)
  {  return (winGetClientSize(hWnd)).height;  }

  public static Rectangle winGetClientSize(HWND hWnd)
  // Retrieves the size of a given window's client area
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetClientSize.htm
  { RECT rect = new RECT();
    AutoItX3.INSTANCE.AU3_WinGetClientSizeByHandle(hWnd, rect);
    return toRectangle(rect);
  }


  public static HWND winGetDesktop()
  // not in AutoIt
  {  return winGetHandle("Program Manager");  }



  public static HWND winGetForeground()
  // get the top-most window (not in AutoIt)
  {  return user32ext.GetTopWindow(null);  }



  public static HWND winGetHandle(String title)
  // retrieves the internal handle of a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetHandle.htm
  {  return AutoItX3.INSTANCE.AU3_WinGetHandle(title, "");  }



  public static Rectangle winGetPos(String title)
  // returns the (x, y, width, height) of a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetPos.htm
  { RECT rect = new RECT();
    AutoItX3.INSTANCE.AU3_WinGetPos(title, "", rect);
    return toRectangle(rect);
  }

  public static Rectangle winGetPos(HWND hWnd)
  { RECT rect = new RECT();
    AutoItX3.INSTANCE.AU3_WinGetPosByHandle(hWnd, rect);
    return toRectangle(rect);
  }



  public static int winGetProcess(String title)
  // Retrieves the Process ID (PID) associated with a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetProcess.htm
  {  return AutoItX3.INSTANCE.AU3_WinGetProcess(title, "");  }

  public static int winGetProcess(HWND hWnd)
  { return AutoItX3.INSTANCE.AU3_WinGetProcessByHandle(hWnd);  }




  public static int winGetState(String title)
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetState.htm
  {  return AutoItX3.INSTANCE.AU3_WinGetState(title, "");  }

  public static int winGetState(HWND hWnd)
  {  return AutoItX3.INSTANCE.AU3_WinGetStateByHandle(hWnd);  }



  public static String winGetText(String title)
  // retrieves the text from a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetText.htm
  { char[] retText = new char[65536];
             // 64kb can be returned == (1024 * 64) = 65536
    AutoItX3.INSTANCE.AU3_WinGetText(title, "", retText, BUF_SIZE);
    return chars2String(retText);
  }

  public static String winGetText(HWND hWnd)
  { char[] retText = new char[65536];
             // 64kb can be returned == (1024 * 64) = 65536
    AutoItX3.INSTANCE.AU3_WinGetTextByHandle(hWnd, retText, BUF_SIZE);
    return chars2String(retText);
  }


  public static String winGetTitle(String title)
  // AutoIt's default mode for matching a window's title is to match from the start;
  // this method returns the full title of the window
  // http://www.autoitscript.com/autoit3/docs/functions/WinGetTitle.htm
  { char[] fullTitle = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_WinGetTitle(title, "", fullTitle, BUF_SIZE);
    return chars2String(fullTitle);
  }

  public static String winGetTitle(HWND hWnd)
  { char[] fullTitle = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_WinGetTitleByHandle(hWnd, fullTitle, BUF_SIZE);
    return chars2String(fullTitle);
  }



  public static int winHide(String title)
  // not in AutoIt
  {  return winSetState(title, SW_HIDE);  }

  public static int winHide(HWND hWnd)
  {  return winSetState(hWnd, SW_HIDE);  }



  public static int winKill(String title)
  // Forces a window to close; useful for overriding default save query
  // http://www.autoitscript.com/autoit3/docs/functions/WinKill.htm
  { System.out.println("Killing \"" + title + "\" (handle: " + 
              handleString(winGetHandle(title)) + ")");
    return AutoItX3.INSTANCE.AU3_WinKill(title, "");  
  }

  public static int winKill(HWND hWnd)
  { System.out.println("Killing \"" + winGetTitle(hWnd) + "\" (handle: " + 
                                              handleString(hWnd) + ")");
    return AutoItX3.INSTANCE.AU3_WinKillByHandle(hWnd);
  }



  public static int winMaximize(String title)
  // not in AutoIt
  {  return winSetState(title, SW_MAXIMIZE);  }

  public static int winMaximize(HWND hWnd)
  {  return winSetState(hWnd, SW_MAXIMIZE);  }


  public static boolean winMaximized(String title)
  {  return ((winGetState(title) & WIN_MAXIMIZED) == WIN_MAXIMIZED);  }

  public static boolean winMaximized(HWND hWnd)
  {  return ((winGetState(hWnd) & WIN_MAXIMIZED) == WIN_MAXIMIZED);  }



  public static boolean winMenuSelectItem(String title, String menuItem)
  // Invokes a menu item of a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinMenuSelectItem.htm
  {  return winMenuSelectItem(title, menuItem, "", "", "", "", "", "", "");  }

  public static boolean winMenuSelectItem(String title, String menuItem, String submenuItem)
  {  return winMenuSelectItem(title, menuItem, submenuItem, "", "", "", "", "", "");  }

  public static boolean winMenuSelectItem(String title,
              String item1, String item2, String item3, String item4, 
              String item5, String item6, String item7, String item8)
  { return (AutoItX3.INSTANCE.AU3_WinMenuSelectItem(title, "", 
              item1, item2, item3, item4, item5, item6, item7, item8) == 1);  }


  public static boolean winMenuSelectItem(HWND hWnd, String menuItem)
  {  return winMenuSelectItem(hWnd, menuItem, "", "", "", "", "", "", "");  }

  public static boolean winMenuSelectItem(HWND hWnd, String menuItem, String submenuItem)
  {  return winMenuSelectItem(hWnd, menuItem, submenuItem, "", "", "", "", "", "");  }

  public static boolean winMenuSelectItem(HWND hWnd,
              String item1, String item2, String item3, String item4,
              String item5, String item6, String item7, String item8)
  { return (AutoItX3.INSTANCE.AU3_WinMenuSelectItemByHandle(hWnd,
              item1, item2, item3, item4, item5, item6, item7, item8) == 1);  }




  public static int winMinimize(String title)
  // not in AutoIt
  {  return winSetState(title, SW_MINIMIZE);  }

  public static int winMinimize(HWND hWnd)
  {  return winSetState(hWnd, SW_MINIMIZE);  }



  public static boolean winMinimized(String title)
  {  return ((winGetState(title) & WIN_MINIMIZED) == WIN_MINIMIZED);  }

  public static boolean winMinimized(HWND hWnd)
  {  return ((winGetState(hWnd) & WIN_MINIMIZED) == WIN_MINIMIZED);  }



  /* I've separated AutoIt's winMove into two separate functions for moving
     and resizing, which are implemented using the 5-arg winMove() */

  public static int winMove(String title, int x, int y)
  { Rectangle rect = winGetPos(title);
    return winMove(title, x, y, rect.width, rect.height);
  }

  public static int winMove(HWND hWnd, int x, int y)
  { Rectangle rect = winGetPos(hWnd);
    return winMove(hWnd, x, y, rect.width, rect.height);
  }


  public static int winMoveToLeftEdge(String title)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(title);
    return winMove(title, scrRect.x, winRect.y, winRect.width, winRect.height);
  }  // end of winMoveToLeftEdge()

  public static int winMoveToRightEdge(String title)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(title);
    int x = scrRect.x + scrRect.width - winRect.width;
    return winMove(title, x, winRect.y, winRect.width, winRect.height);
  }  // end of winMoveToRightEdge()

  public static int winMoveToTopEdge(String title)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(title);
    return winMove(title, winRect.x, scrRect.y, winRect.width, winRect.height);
  }  // end of winMoveToTopEdge()

  public static int winMoveToBottomEdge(String title)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(title);
    int y = scrRect.y + scrRect.height - winRect.height;
    return winMove(title, winRect.x, y, winRect.width, winRect.height);
  }  // end of winMoveToBottomEdge()

  public static int winMoveToCenter(String title)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(title);
    int x = (scrRect.x + scrRect.width - winRect.width)/2;
    int y = (scrRect.y + scrRect.height - winRect.height)/2;
    return winMove(title, x, y, winRect.width, winRect.height);
  }  // end of winMoveToCenter()



  public static int winMoveToLeftEdge(HWND hWnd)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(hWnd);
    return winMove(hWnd, scrRect.x, winRect.y, winRect.width, winRect.height);
  }  // end of winMoveToLeftEdge()

  public static int winMoveToRightEdge(HWND hWnd)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(hWnd);
    int x = scrRect.x + scrRect.width - winRect.width;
    return winMove(hWnd, x, winRect.y, winRect.width, winRect.height);
  }  // end of winMoveToRightEdge()

  public static int winMoveToTopEdge(HWND hWnd)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(hWnd);
    return winMove(hWnd, winRect.x, scrRect.y, winRect.width, winRect.height);
  }  // end of winMoveToTopEdge()

  public static int winMoveToBottomEdge(HWND hWnd)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(hWnd);
    int y = scrRect.y + scrRect.height - winRect.height;
    return winMove(hWnd, winRect.x, y, winRect.width, winRect.height);
  }  // end of winMoveToBottomEdge()

  public static int winMoveToCenter(HWND hWnd)
  { Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(hWnd);
    int x = (scrRect.x + scrRect.width - winRect.width)/2;
    int y = (scrRect.y + scrRect.height - winRect.height)/2;
    return winMove(hWnd, x, y, winRect.width, winRect.height);
  }  // end of winMoveToCenter()



  public static int winMove(String title, int x, int y, int width, int height)
  // Moves and/or resizes a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinMove.htm
  // the AutoItX function does not have a speed argument
  {  
    Rectangle scrRect = getScreenBounds();

    if ((x+width) < scrRect.x) {
      System.out.println("Error: window will be off left side of the screen");
      x = scrRect.x - width + 10;
    }
    else if (x > (scrRect.x + scrRect.width)) {
      System.out.println("Error: window will be off right side of the screen");
      x = scrRect.x + scrRect.width - 10;
    }

    if ((y+height) < scrRect.y) {
      System.out.println("Error: window will be off top of the screen");
      y = scrRect.y - height + 10;
    }
    else if (y > (scrRect.y + scrRect.height)) {
      System.out.println("Error: window will be off bottom of the screen");
      y = scrRect.y + scrRect.height - 10;
    }

    return AutoItX3.INSTANCE.AU3_WinMove(title, "", x, y, width, height);  
  }  // end of winMove()



  public static int winMove(HWND hWnd, int x, int y, int width, int height)
  // Moves and/or resizes a window
  {  
    Rectangle scrRect = getScreenBounds();

    if ((x+width) < scrRect.x) {
      System.out.println("Error: window will be off left side of the screen");
      x = scrRect.x - width + 10;
    }
    else if (x > (scrRect.x + scrRect.width)) {
      System.out.println("Error: window will be off right side of the screen");
      x = scrRect.x + scrRect.width - 10;
    }

    if ((y+height) < scrRect.y) {
      System.out.println("Error: window will be off top of the screen");
      y = scrRect.y - height + 10;
    }
    else if (y > (scrRect.y + scrRect.height)) {
      System.out.println("Error: window will be off bottom of the screen");
      y = scrRect.y + scrRect.height - 10;
    }

    return AutoItX3.INSTANCE.AU3_WinMoveByHandle(hWnd, x, y, width, height);  
  }  // end of winMove()




  public static int winResize(String title, int width, int height)
  { Rectangle winRect = winGetPos(title);
    return winMove(title, winRect.x, winRect.y, width, height);
  }

  public static int winResize(HWND hWnd, int width, int height)
  { Rectangle winRect = winGetPos(hWnd);
    return winMove(hWnd, winRect.x, winRect.y, width, height);
  }


  public static int winResize(String title, double screenFraction)
  { 
    Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(title);
    screenFraction = checkScrFrac(screenFraction);
    int width = (int) Math.round(scrRect.width*screenFraction);
    int height = (int) Math.round(scrRect.height*screenFraction);
    return winMove(title, winRect.x, winRect.y, width, height);
  }  // end of winResize()


  public static int winResize(HWND hWnd, double screenFraction)
  { 
    Rectangle scrRect = getScreenBounds();
    Rectangle winRect = winGetPos(hWnd);
    screenFraction = checkScrFrac(screenFraction);
    int width = (int) Math.round(scrRect.width*screenFraction);
    int height = (int) Math.round(scrRect.height*screenFraction);
    return winMove(hWnd, winRect.x, winRect.y, width, height);
  }  // end of winResize()


  private static double checkScrFrac(double screenFraction)
  {
    if (screenFraction < 0.2) {
      System.out.println("Resize fraction too small; using 0.2");
      screenFraction = 0.2;
    }
    else if (screenFraction > 1.0) {
      System.out.println("Resize fraction too big; using 1.0");
      screenFraction = 1.0;
    }
    return screenFraction;
  }  // end of checkScrFrac()




  public static int winRestore(String title)
  // Undoes a window minimization or maximization (not in AutoIt)
  {  winSetState(title, SW_RESTORE);  
     return winActivate(title);
  }

  public static int winRestore(HWND hWnd)
  {  winSetState(hWnd, SW_RESTORE);  
     return winActivate(hWnd);
  }


  public static boolean winSetOnTop(String title, boolean onTop)
  // Change a window's "Always On Top" attribute
  // http://www.autoitscript.com/autoit3/docs/functions/WinSetOnTop.htm
  { int flag = (onTop) ? 1 : 0;
     return (AutoItX3.INSTANCE.AU3_WinSetOnTop(title, "", flag) == 1);
  }

  public static boolean winSetOnTop(HWND hWnd, boolean onTop)
  { int flag = (onTop) ? 1 : 0;
     return (AutoItX3.INSTANCE.AU3_WinSetOnTopByHandle(hWnd, flag) == 1);
  }




  public static int winSetState(String title, int flag)
  // Shows, hides, minimizes, maximizes, or restores a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinSetState.htm
  {  return AutoItX3.INSTANCE.AU3_WinSetState(title, "", flag);  }

  public static int winSetState(HWND hWnd, int flag)
  {  return AutoItX3.INSTANCE.AU3_WinSetStateByHandle(hWnd, flag);  }



  public static int winSetTitle(String title, String newTitle)
  // Changes the title of a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinSetTitle.htm
  {  return AutoItX3.INSTANCE.AU3_WinSetTitle(title, "", newTitle);  }

  public static int winSetTitle(HWND hWnd, String newTitle)
  {  return AutoItX3.INSTANCE.AU3_WinSetTitleByHandle(hWnd, newTitle);  }



  public static int winSetVisibility(String title, boolean isVisible)
  // not in AutoIt
  { int transparency = (isVisible) ? 255 : 0;
         /* A number in the range 0 - 255. The lower the number,
            the more transparent the window will become. */
    return winSetTrans(title, transparency);
  }

  public static int winSetVisibility(HWND hWnd, boolean isVisible)
  { int transparency = (isVisible) ? 255 : 0;
    return winSetTrans(hWnd, transparency);
  }



  public static int winSetTrans(String title, double transp)
  // Sets the transparency of a window
  // http://www.autoitscript.com/autoit3/docs/functions/WinSetTrans.htm
  {
    int transparency = (int) Math.round(transp*255);
    if (transparency > 255)
      transparency = 255;
    else if (transparency < 0)
      transparency = 0;
    return AutoItX3.INSTANCE.AU3_WinSetTrans(title, "", transparency);
  }  // end of winSetTrans()


  public static int winSetTrans(HWND hWnd, double transp)
  // Sets the transparency of a window
  {
    int transparency = (int) Math.round(transp*255);
    if (transparency > 255)
      transparency = 255;
    else if (transparency < 0)
      transparency = 0;
    return AutoItX3.INSTANCE.AU3_WinSetTransByHandle(hWnd, transparency);
  }  // end of winSetTrans()



  public static int winShow(String title)
  // Shows a previously hidden window
  {  return winSetState(title, SW_SHOW);  }

  public static int winShow(HWND hWnd)
  {  return winSetState(hWnd, SW_SHOW);  }




  public static int winTitleMatchMode(int flag)
  /* Alters the method that is used to match window titles during search operations.
     flag values: 
       1 = (default) Match the title from the start
       2 = Match any substring in the title
       3 = Exact title match
       4 = Advanced mode, see Window Titles & Text (Advanced)
       -1 to -4 = Case insensitive match according to the other type of match.
    see http://www.autoitscript.com/autoit3/docs/functions/AutoItSetOption.htm
  */
  {  
    if ((flag >= -4) && (flag <= 4) && (flag != 0)) {
      System.out.println("Changing title match mode to " + flag);
      return autoItSetOption("WinTitleMatchMode", flag);
    }
    else {
      System.out.println("Illegal title match mode value: " + flag);
      return 0;
    }
  }  // end of winTitleMatchMode()


  public static boolean winToBack(String title)
  {  return winToBack( winGetHandle(title));  }


  public static boolean winToBack(HWND hWnd)
  {
    user32ext.SetWindowPos(hWnd, user32ext.HWND_BOTTOM, 0, 0, 0, 0, 
                   user32ext.SWP_NOMOVE | user32ext.SWP_NOSIZE);
    return true;
  }  // end of winToBack()




  public static boolean winToForeground(String title)
  {  return user32ext.SetForegroundWindow( winGetHandle(title) );   }

  public static boolean winToForeground(HWND hWnd)
  {  return user32ext.SetForegroundWindow(hWnd);   }





  public static boolean winVisible(String title)
  {  return ((winGetState(title) & WIN_VISIBLE) == WIN_VISIBLE);  }

  public static boolean winVisible(HWND hWnd)
  {  return ((winGetState(hWnd) & WIN_VISIBLE) == WIN_VISIBLE);  }



  /* the AutoITx headers for winWait() and winWaitActive() return an int 
     not a handle, unlike as stated in the function docs */


  public static HWND winWaitHandle(String title)
  {  boolean isActive = winWait(title, 0);  
     return ((isActive) ? winGetHandle(title) : null);
  }

  public static HWND winWaitHandle(String title, int timeOut)
  {  boolean isActive = winWait(title, timeOut);
     return ((isActive) ? winGetHandle(title) : null);
  }


  public static boolean winWait(String title)
  // pauses execution of the script until the requested window exists
  // http://www.autoitscript.com/autoit3/docs/functions/WinWait.htm
  {  return winWait(title, 0);  }

  public static boolean winWait(String title, int timeOut)
  {  return (AutoItX3.INSTANCE.AU3_WinWait(title, "", timeOut) == 1); }

  public static boolean winWait(HWND hWnd)
  {  return winWait(hWnd, 0);  }

  public static boolean winWait(HWND hWnd, int timeOut)
  {  return (AutoItX3.INSTANCE.AU3_WinWaitByHandle(hWnd, timeOut) == 1); }




  public static HWND winWaitActiveHandle(String title)
  {  boolean isActive = winWaitActive(title, 0);  
     return ((isActive) ? winGetHandle(title) : null);
  }

  public static HWND winWaitActiveHandle(String title, int timeOut)
  {  boolean isActive = winWaitActive(title, timeOut);
     return ((isActive) ? winGetHandle(title) : null);
  }


  public static boolean winWaitActive(String title)
  // pauses execution of the script until the requested window is active
  // http://www.autoitscript.com/autoit3/docs/functions/WinWaitActive.htm
  {  return winWaitActive(title, 0);  }

  public static boolean winWaitActive(String title, int timeOut)
  {  return (AutoItX3.INSTANCE.AU3_WinWaitActive(title, "", timeOut) == 1);  }

  public static boolean winWaitActive(HWND hWnd)
  {  return winWaitActive(hWnd, 0);  }

  public static boolean winWaitActive(HWND hWnd, int timeOut)
  {  return (AutoItX3.INSTANCE.AU3_WinWaitActiveByHandle(hWnd, timeOut) == 1);  }



  public static boolean winWaitClose(String title)
  // Pauses execution of the script until the requested window does not exist
  // http://www.autoitscript.com/autoit3/docs/functions/WinWaitClose.htm
  {  return winWaitClose(title, 0);  }

  public static boolean winWaitClose(String title, int timeOut)
  {  
    System.out.println("Closing \"" + title + "\" (handle: " + 
                                    handleString(winGetHandle(title)) + ")");
    return (AutoItX3.INSTANCE.AU3_WinWaitClose(title, "", timeOut) == 1);
  }

  public static boolean winWaitClose(HWND hWnd)
  {  return winWaitClose(hWnd, 0);  }

  public static boolean winWaitClose(HWND hWnd, int timeOut)
  { System.out.println("Closing \"" + winGetTitle(hWnd) + "\" (handle: " + handleString(hWnd) + ")");
    return (AutoItX3.INSTANCE.AU3_WinWaitCloseByHandle(hWnd, timeOut) == 1);
  }



  public static boolean winWaitNotActive(String title)
  // Pauses execution of the script until the requested window is not active
  // http://www.autoitscript.com/autoit3/docs/functions/WinWaitNotActive.htm
  {  return winWaitNotActive(title, 0);  }

  public static boolean winWaitNotActive(String title, int timeOut)
  {  return (AutoItX3.INSTANCE.AU3_WinWaitNotActive(title, "", timeOut) == 1);  }

  public static boolean winWaitNotActive(HWND hWnd)
  {  return winWaitNotActive(hWnd, 0);  }

  public static boolean winWaitNotActive(HWND hWnd, int timeOut)
  {  return (AutoItX3.INSTANCE.AU3_WinWaitNotActiveByHandle(hWnd, timeOut) == 1);  }



  public static Rectangle getScreenBounds()
  // http://stackoverflow.com/questions/1936566/how-do-you-get-the-screen-width-in-java
  // uses the default monitor only
  {
    java.awt.GraphicsConfiguration gc =
         java.awt.GraphicsEnvironment
           .getLocalGraphicsEnvironment()
           .getDefaultScreenDevice()
           .getDefaultConfiguration();

    Rectangle scrBounds = gc.getBounds();
    Insets si = Toolkit.getDefaultToolkit().getScreenInsets(gc);

    scrBounds.x += si.left;
    scrBounds.y += si.top;
    scrBounds.width -= si.left + si.right;
    scrBounds.height-= si.top + si.bottom;

    return scrBounds;
  }  // end of getScreenBounds()



  public static Rectangle toRectangle(RECT rect)
  { int width = rect.right - rect.left;
    int height = rect.bottom - rect.top;
    return new Rectangle(rect.left, rect.top, width, height);
  }  // end of toRectangle()


  public static RECT toRECT(Rectangle rect)
  {
    RECT lpRect = new RECT();
    lpRect.left = rect.x;
    lpRect.top = rect.y;
    lpRect.right = rect.x + rect.width;
    lpRect.bottom = rect.y + rect.height;
    return lpRect;
  }  // end of toRECT()



  //------------------------------ send message methods -----------------------------

  public static void send(String sendText)
  {  send(sendText, 0);  }


  public static void send(String sendText, int mode)
  // Sends simulated keystrokes to the active window
  /* mode: Changes how "keys" is processed:
       mode == 0 (default): Text contains special chars like + and !
                            to indicate SHIFT and ALT key-presses.
       mode == 1: keys are sent raw.
    See http://www.autoitscript.com/autoit3/docs/functions/Send.htm
  */
  {   AutoItX3.INSTANCE.AU3_Send(sendText, mode);  }


  public static void send(HWND hWnd, String sendText)
  // change the active window before sending keystrokes
  { winActivate(hWnd);
    send(sendText);
  }


  public static void send(HWND hWnd, String sendText, int mode)
  { winActivate(hWnd);
    send(sendText, mode);
  }


  //------------------------- controls ----------------------------------




  public static boolean controlClick(String title, int ctrlID)
  {  return controlClick(title, ctrlID, BTN_LEFT);  }


  public static boolean controlClick(String title, int ctrlID, String button)
  {  return controlClick(title, ctrlID, button, 1,  
                                      AutoItX3.AU3_INTDEFAULT, AutoItX3.AU3_INTDEFAULT);
  }


  public static boolean controlClick(String title, int ctrlID,
                             String button, int numClicks, int x, int y)
  /* sends mouse click(s) to a given control
         button: the button to click, "left", "right", "middle", "main", "menu", 
                 "primary", "secondary". Default is the left button.
         numClicks: the number of times to click the mouse. Default is 1.
         x: the x position to click within the control. Default is center of control.
         y: the y position to click within the control. Default is center.

     see http://www.autoitscript.com/autoit3/docs/functions/ControlClick.htm
  */
  { if (!isMouseButton(button))
      return false;
    return (AutoItX3.INSTANCE.AU3_ControlClick(title, "", controlIDDescriptor(ctrlID), 
                                                             button,  numClicks, x, y) == 1);
  }

  public static boolean controlClick(HWND hWnd, int ctrlID)
  {  return controlClick(hWnd, ctrlID, BTN_LEFT);  }


  public static boolean controlClick(HWND hWnd, int ctrlID, String button)
  {  return controlClick(hWnd, ctrlID, button, 1, AutoItX3.AU3_INTDEFAULT, AutoItX3.AU3_INTDEFAULT);
  }


  public static boolean controlClick(HWND hWnd, int ctrlID,
                                   String button, int numClicks, int x, int y)
  { HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    return (AutoItX3.INSTANCE.AU3_ControlClickByHandle(hWnd, hCtrl, button, numClicks, x, y) == 1);
  }





  public static String controlCommand(String title, int ctrlID, String cmd, String option)
  // Sends a command to a control
  // see https://www.autoitscript.com/autoit3/docs/functions/ControlCommand.htm
  { char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlCommand(title, "", controlIDDescriptor(ctrlID), 
                                                   cmd, option, result, BUF_SIZE);
    return chars2String(result);
  }


  public static String controlCommand(HWND hWnd, int ctrlID, String cmd, String option)
  { 
    HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlCommandByHandle(hWnd, hCtrl, cmd, option, result, BUF_SIZE);
    return chars2String(result);
  }



  public static boolean controlFocus(String title, int ctrlID)
  // sets input focus to a given control on a window
  // http://www.autoitscript.com/autoit3/docs/functions/ControlFocus.htm
  {  return (AutoItX3.INSTANCE.AU3_ControlFocus(title, "", controlIDDescriptor(ctrlID)) == 1);  }

  public static boolean controlFocus(HWND hWnd, int ctrlID)
  { HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    return (AutoItX3.INSTANCE.AU3_ControlFocusByHandle(hWnd, hCtrl) == 1);
  }



  public static int controlGetFocus(String title)
  // Returns the handle of the control that has keyboard focus within a specified window
  // http://www.autoitscript.com/autoit3/docs/functions/ControlGetFocus.htm
  //  modified to return the control ID rather than the ClassNameNN string
  { 
    char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlGetFocus(title, "", result, BUF_SIZE);
    String classNN = chars2String(result);   // the ControlRef# of the control
    HWND hCtrl = AutoItX3.INSTANCE.AU3_ControlGetHandle( winGetHandle(title), classNN);
           // the handle of the control
    return controlIDFromHandle(hCtrl);
  }

  public static int controlGetFocus(HWND hWnd)
  { char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlGetFocusByHandle(hWnd, result, BUF_SIZE);
    String classNN = chars2String(result);
    HWND hCtrl = AutoItX3.INSTANCE.AU3_ControlGetHandle(hWnd, classNN);
           // the handle of the control
    return controlIDFromHandle(hCtrl);
  }

  public static String controlGetClassName(String title, int ctrlID)
  { HWND hWnd = winGetHandle(title);
    return controlGetClassName(hWnd, ctrlID);
  }

  public static String controlGetClassName(HWND hWnd, int ctrlID)
  // get the control's class name, e.g. "Button", "Static"
  // see http://www.autoitscript.com/autoit3/docs/intro/controls.htm
  // this function isn't listed in the AutoIt functions list
  {
    HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    char[] textBuffer = new char[512];
    user32ext.GetClassName(hCtrl, textBuffer, 512);
    return (new String(textBuffer)).trim();
  }  // end of controlGetClassName()



  public static HWND controlGetHandle(String title, int ctrlID)
  // retrieves the internal handle of a control
  // http://www.autoitscript.com/autoit3/docs/functions/ControlGetHandle.htm
  {  return AutoItX3.INSTANCE.AU3_ControlGetHandle(winGetHandle(title), 
                                                   controlIDDescriptor(ctrlID));
  }

  public static HWND controlGetHandle(HWND hWnd, int ctrlID)
  // retrieves the internal handle of a control
  {  return AutoItX3.INSTANCE.AU3_ControlGetHandle(hWnd, controlIDDescriptor(ctrlID));  }



  public static String controlIDDescriptor(int ctrlID)
  // not in AutoIt
  {  return ("[ID:" + ctrlID + "]");  }




  public static List<Integer> controlIDsList(String title)
  // return a list of control IDs for the window title (not in AutoIt)
  {  return controlIDsList(winGetHandle(title));  }


  public static List<Integer> controlIDsList(final HWND hWnd)
  // return a list of control IDs for the window handle (not in AutoIt)
  {
    // System.out.println("getting controls");
    final List<Integer> ctrlIDs = new ArrayList<Integer>();

    // now call EnumChildWindows() using handle of the window
    user32ext.EnumChildWindows(hWnd, new WNDENUMPROC()
    {
      public boolean callback(HWND hCtrl, Pointer userData)
      { // called for each control that EnumChildWindows() finds
        int ctrlID = controlIDFromHandle(hCtrl);
        String cNm = controlGetClassName(hWnd, ctrlID);
        if (!cNm.isEmpty())
          ctrlIDs.add(ctrlID);
        return true;
      }
    }, null);

    // System.out.println("finished getting controls");
    return ctrlIDs;
  }  // end of controlIDsList()



  public static int controlIDFromHandle(HWND hCtrl)
  // return the control ID for the control handle (not part of AutoIt)
  {  return user32ext.GetDlgCtrlID(hCtrl);  }



/*
  public static int controlGetPosX(String title, int ctrlID)
  {  return (controlGetPos(title, ctrlID)).x;  }

  public static int controlGetPosY(String title, int ctrlID)
  {  return (controlGetPos(title, ctrlID)).y;  }

  public static int controlGetPosWidth(String title, int ctrlID)
  {  return (controlGetPos(title, ctrlID)).width;  }

  public static int controlGetPosHeight(String title, int ctrlID)
  {  return (controlGetPos(title, ctrlID)).height;  }

  public static int controlGetPosX(HWND hWnd, int ctrlID)
  {  return (controlGetPos(hWnd, ctrlID)).x;  }

  public static int controlGetPosY(HWND hWnd, int ctrlID)
  {  return (controlGetPos(hWnd, ctrlID)).y;  }

  public static int controlGetPosWidth(HWND hWnd, int ctrlID)
  {  return (controlGetPos(hWnd, ctrlID)).width;  }

  public static int controlGetPosHeight(HWND hWnd, int ctrlID)
  {  return (controlGetPos(hWnd, ctrlID)).height;  }
*/

  public static Rectangle controlGetPos(String title, int ctrlID)
  // Retrieves the position and size of a control relative to its window
  // http://www.autoitscript.com/autoit3/docs/functions/ControlGetPos.htm
  {
    RECT rect = new RECT();
    AutoItX3.INSTANCE.AU3_ControlGetPos(title, "", controlIDDescriptor(ctrlID), rect);
    return toRectangle(rect);
  }

  public static Rectangle controlGetPos(HWND hWnd, int ctrlID)
  { HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    RECT rect = new RECT();
    AutoItX3.INSTANCE.AU3_ControlGetPosByHandle(hWnd, hCtrl, rect);
    return toRectangle(rect);
  }



  public static String controlGetText(String title, int ctrlID)
  // Retrieves text from a control.
  // http://www.autoitscript.com/autoit3/docs/functions/ControlGetText.htm
  { char[] controlGetText = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlGetText(title, "", controlIDDescriptor(ctrlID), 
                                                            controlGetText, BUF_SIZE);
    return chars2String(controlGetText);
  }


  public static String controlGetText(HWND hWnd, int ctrlID)
  { HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    char[] controlGetText = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlGetTextByHandle(hWnd, hCtrl, controlGetText, BUF_SIZE);
    return chars2String(controlGetText);
  }




  public static String controlListView(String title, int ctrlID,
                                      String cmd, String option1, String option2)
  // Sends a command to a ListView32 control.
  // http://www.autoitscript.com/autoit3/docs/functions/ControlListView.htm
  { char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlListView(title, "", controlIDDescriptor(ctrlID), cmd,
                                          option1, option2, result, BUF_SIZE);
    return chars2String(result);
  }


  public static String controlListView(HWND hWnd, int ctrlID,
                                      String cmd, String option1, String option2)
  { HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlListViewByHandle(hWnd, hCtrl, cmd,
                                       option1, option2, result, BUF_SIZE);
    return chars2String(result);
  }




  public static boolean controlSend(String title, int ctrlID, String sendText)
  {  return controlSend(title, ctrlID, sendText, 0);  }


  public static boolean controlSend(String title, int ctrlID, String sendText, int mode)
  {  return (AutoItX3.INSTANCE.AU3_ControlSend(title, "", 
                                 controlIDDescriptor(ctrlID), sendText, mode) == 1);  }


  public static boolean controlSend(HWND hWnd, int ctrlID, String sendText)
  {  return controlSend(hWnd, ctrlID, sendText, 0);  }


  public static boolean controlSend(HWND hWnd, int ctrlID, String sendText, int mode)
  /* sends a String of characters (msg) to a control
       mode: Changes how the characters in the text are processed:
          mode == 0 (default), text contains special characters like + to indicate
                               SHIFT and {LEFT} to indicate left arrow.
          mode == 1, keys are sent raw. 
     see http://www.autoitscript.com/autoit3/docs/functions/ControlSend.htm
  */
  {  HWND hCtrl = controlGetHandle(hWnd, ctrlID);
     return (AutoItX3.INSTANCE.AU3_ControlSendByHandle(hWnd, hCtrl, sendText, mode) == 1);
  }




  public static boolean controlSetText(String title, int ctrlID, String txt)
  // Sets text of a control.
  // http://www.autoitscript.com/autoit3/docs/functions/ControlSetText.htm
  {  return (AutoItX3.INSTANCE.AU3_ControlSetText(title, "", controlIDDescriptor(ctrlID), txt) == 1);  }


  public static boolean controlSetText(HWND hWnd, int ctrlID, String txt)
  { HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    return (AutoItX3.INSTANCE.AU3_ControlSetTextByHandle(hWnd, hCtrl, txt) == 1);
  }




  public static String controlTreeView(String title, int ctrlID, String cmd, String option1)
  // Sends a command to a TreeView32 control
  // http://www.autoitscript.com/autoit3/docs/functions/ControlTreeView.htm
  { char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlTreeView(title, "", controlIDDescriptor(ctrlID), cmd,
                                          option1, "", result, BUF_SIZE);
    return chars2String(result);
  }

  public static String controlTreeView(HWND hWnd, int ctrlID, String cmd, String option1)
  { HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    char[] result = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_ControlTreeViewByHandle(hWnd, hCtrl, cmd,
                                           option1, "", result, BUF_SIZE);
    return chars2String(result);
  }





  // ---------------------- clip board -----------------------------------



  public static String clipGet(int timeOut)
  // keep polling the clipboard for up to timeOut ms looking for text
  // not in AutoIt
  {
    // System.out.println("Waiting for " + timeOut + " ms for window to open...");
    int counter = 0;
    String clipText = clipGet();
    while ((clipText == null) && (counter < timeOut)) {
      sleep(1000);
      clipText = clipGet();
      counter += 1000;
    }

    if (clipText == null)
      System.out.println("No cliptext found before timeOut");

    return clipText;
  }  // end of clipGet()



  public static String clipGet()
  // http://www.autoitscript.com/autoit3/docs/functions/ClipGet.htm
  /* Or use
          Clipboard cb = Toolkit.getDefaultToolkit().getSystemClipboard();
  */
  {
    char[] result = new char[65536];
    AutoItX3.INSTANCE.AU3_ClipGet(result, 65536);
    return chars2String(result);
  }



  public static boolean clipPut(String clip)
  // http://www.autoitscript.com/autoit3/docs/functions/ClipPut.htm
  {   return ( AutoItX3.INSTANCE.AU3_ClipPut(clip) == 1);  }


  public static boolean clipClear()
  {  return clipPut("");  }


  //-----------------------status bar -----------------------------


  public static String statusbarGetText(String title)
  // Retrieves the text from a standard status bar control
  // http://www.autoitscript.com/autoit3/docs/functions/StatusbarGetText.htm
  {  return statusbarGetText(title, 1);  }

  public static String statusbarGetText(String title, int part)
  { char[] statusText = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_StatusbarGetText(title, "", part, statusText, BUF_SIZE);
    return chars2String(statusText);
  }


  public static String statusbarGetText(HWND hWnd, int ctrlID)
  {  return statusbarGetText(hWnd, ctrlID, 1);  }

  public static String statusbarGetText(HWND hWnd, int ctrlID, int part)
  { 
    HWND hCtrl = controlGetHandle(hWnd, ctrlID);
    char[] statusText = new char[BUF_SIZE];
    AutoItX3.INSTANCE.AU3_StatusbarGetTextByHandle(hCtrl, part, statusText, BUF_SIZE);
    return chars2String(statusText);
  }




 //----------------------------- mouse methods ----------------------------



  public static int mouseCoordMode(int flag)
  /* Sets the way coords are used in the mouse functions, either absolute coords 
     or coords relative to the current active window:
         0 (or M_WINDOW) = relative coords to the active window
         1 (or M_SCREEN) = (default) absolute screen coordinates
         2 (or M_CLIENT)= relative coords to the client area of the active window
    see http://www.autoitscript.com/autoit3/docs/functions/AutoItSetOption.htm
  */
  {  
    if ((flag >= 0) && (flag <= 3)) {
      System.out.println("Changing mouse coord mode to " + flag);
      return autoItSetOption("MouseCoordMode", flag);
    }
    else {
      System.out.println("Illegal mouse coord mode value: " + flag);
      return 0;
    }
  }  // end of mouseCoordMode()





  public static boolean mouseClick(String button)
  { Point pos = mouseGetPos();
    return mouseClick(button, pos.x, pos.y);
  }


  public static boolean mouseClick(String button, int x, int y)
  {  return mouseClick(button, x, y, 1, 10);   }


  public static boolean mouseClick(String button, int x, int y, int numClicks, int speed)
  /* Perform a mouse click
        button: the button to click: "left", "right", "middle", "main", "menu",
                                      "primary", "secondary".
        numClicks: the number of times to click the mouse. Default is 1.
        speed: the speed to move the mouse in the range 1 (fastest) to 100 (slowest).
               A speed of 0 will move the mouse instantly. Default speed is 10.

    See http://www.autoitscript.com/autoit3/docs/functions/MouseClick.htm
  */
  { if (!isMouseButton(button))
      return false;
    else
     return (AutoItX3.INSTANCE.AU3_MouseClick(button, x, y, numClicks, speed) == 1);
  }



  public static boolean mouseClickDrag(String button, int x1, int y1, int x2, int y2)
  {  return mouseClickDrag(button, x1, y1, x2, y2, 10);  }


  public static boolean mouseClickDrag(String button, int x1, int y1, int x2, int y2, int speed)
  // Perform a mouse click and drag from (x1,y1) to (x2,y2)
  // http://www.autoitscript.com/autoit3/docs/functions/MouseClickDrag.htm
  { if (!isMouseButton(button))
      return false;
    else
      return (AutoItX3.INSTANCE.AU3_MouseClickDrag(button, x1, y1, x2, y2, speed) == 1);
  }


  public static boolean mouseDown(String button)
  // Perform a mouse down event at the current mouse position
  // http://www.autoitscript.com/autoit3/docs/functions/MouseDown.htm
  { if (!isMouseButton(button))
      return false;
    else
      return (AutoItX3.INSTANCE.AU3_MouseDown(button) == 1);
  }


  public static int mouseGetCursorID()
  // Returns the cursor ID Number for the current Mouse Cursor
  // http://www.autoitscript.com/autoit3/docs/functions/MouseGetCursor.htm
  {  return AutoItX3.INSTANCE.AU3_MouseGetCursor();  }



  public static Point mouseGetPos()
  // return the current (x, y) position of the mouse cursor
  // http://www.autoitscript.com/autoit3/docs/functions/MouseGetPos.htm
  { WinDef.POINT point = new WinDef.POINT();
    AutoItX3.INSTANCE.AU3_MouseGetPos(point);
    return new Point(point.x, point.y);
  }



  public static void mouseMoveBy(int xStep, int yStep)
  { Point pos = mouseGetPos();
    mouseMoveTo(pos.x+xStep, pos.y+yStep, 10);  
  }

  public static void mouseMoveBy(int xStep, int yStep, int speed)
  { Point pos = mouseGetPos();
    mouseMoveTo(pos.x+xStep, pos.y+yStep, speed);
  }



  public static void mouseMoveTo(int x, int y)
  // http://www.autoitscript.com/autoit3/docs/functions/MouseMove.htm
  {  mouseMoveTo(x, y, 10);  }

  public static void mouseMoveTo(int x, int y, int speed)
  {  AutoItX3.INSTANCE.AU3_MouseMove(x, y, speed);  }



  public static boolean mouseUp(String button)
  // Perform a mouse up event at the current mouse position
  // http://www.autoitscript.com/autoit3/docs/functions/MouseUp.htm
  { if (!isMouseButton(button))
      return false;
    else
      return (AutoItX3.INSTANCE.AU3_MouseUp(button) == 1);
  }


  public static boolean mouseWheel(String wheelDir)
  // Moves the mouse wheel up or down
  // http://www.autoitscript.com/autoit3/docs/functions/MouseWheel.htm
  {  return mouseWheel(wheelDir, 1);  }


  public static boolean mouseWheel(String wheelDir, int numClicks)
  { if (!isWheelDir(wheelDir))
      return false;
    else
      return (AutoItX3.INSTANCE.AU3_MouseWheel(wheelDir, numClicks) == 1);
  }


  public static HWND mouseWindowHandle()
  // gets handle of top-most window below the current mouse position (not in AutoIt)
  {
    WinDef.POINT point = new WinDef.POINT();
    AutoItX3.INSTANCE.AU3_MouseGetPos(point);
    // System.out.println("(" + point.x + ", " + point.y + ")");
    return user32ext.WindowFromPoint(point.x, point.y);
  }


  public static boolean isMouseButton(String button)
  {
    for (String b : BTNS) {
      if (button.equals(b))
        return true;
    }
    System.out.println(button + " is not a recognized button label");
    return false;
  }  // end of isMouseButton()


  public static boolean isWheelDir(String wheelDir)
  {
    for (String dir : DIRS) {
      if (wheelDir.equals(dir))
        return true;
    }
    System.out.println(wheelDir + " is not a recognized wheel direction");
    return false;
  }  // end of isWheelDir()




  // ---------------------- support methods ----------------------

  public static String getAutoItDLL()
  // return the path to the 32-bit or 64-bit version of the AutoIT DLL
  // used by AutoItX3.java
  {

    if (!System.getProperty("os.name").contains("Windows")) {
      System.out.println("OS is not Windows");
       return null;
    }

    String dllPath = null;
    String programFiles = System.getenv("ProgramFiles(x86)");
    // System.out.println("PF (x86): " + programFiles);
    // System.out.println("PF: " + System.getenv("ProgramFiles"));

    if (programFiles == null)   // must be a 32-bit version of Windows
      dllPath = System.getenv("ProgramFiles") + "\\AutoIt3\\AutoItX\\AutoItX3.dll";
    else   // must be a 64-bit version
      dllPath = programFiles + "\\AutoIt3\\AutoItX\\AutoItX3_x64.dll";

    // System.out.println("Assuming the AutoIt DLL is located at \"" + dllPath + "\"");
    return dllPath;
  }  // end of getAutoItDLL()



  public static String chars2String(char[] msg)
  // change a char array to an ASCII string, with extra spaces trimmed
  {
    int numChanges = 0;
    for (int i=0; i < msg.length; i++)
       if (msg[i] > 127) {    // ASCII is in range 0 to 127
         msg[i] = ' ';
         numChanges++;
       }
     // System.out.println("Num changes: " + numChanges);
     String msgStr = new String(msg).trim();
     if (msgStr.equals("")) {
       // System.out.println("array contains no text");
       return null;
     }
     else
       return msgStr;
  }  // end of chars2String()



  public static void msgBox(String title, String text)
  // Java version of AutoIt's MsgBox()
  {  JOptionPane.showMessageDialog(null, title, text, JOptionPane.INFORMATION_MESSAGE);  }



  public static int toInt(String valStr)
  { try {
      return Integer.parseInt(valStr);
    }
    catch(NumberFormatException e)
    { System.out.println(valStr + " is not an integer");  }
    return -1;
  }  // end of toInt()



  public static String hexString(int val) 
  // convert the number to a 4-digit hexadecimal string;
  // useful for pretty-printing control IDs
  {  return String.format("0x%4s", Integer.toHexString(val)).replace(' ', '0');   }



  public static void sleep(int ms)
  // Java replacement for AutoIt sleep()
  { try {
		  Thread.sleep(ms);
	  } 
    catch (InterruptedException e) {}
  }  // end of sleep()


  public static boolean isWindows()
  { String OS = System.getProperty("os.name").toLowerCase();
    return (OS.indexOf("win") >= 0);
  }


  public static String windowsVersion()
  { if (!isWindows())
      return null;
    else {
      return ( System.getProperty("os.name") + "; " + 
               System.getProperty("os.version") );
    }
  }  // end of windowsVersion()

}  // end of Jau class
